@@ -451,6 +451,17 @@
 #define PCI_CHIP_RS780_9614 0x9614
 #define PCI_CHIP_RS780_9615 0x9615
 #define PCI_CHIP_RS780_9616 0x9616
+#define PCI_CHIP_SUMO_9640 0x9640
+#define PCI_CHIP_SUMO_9641 0x9641
+#define PCI_CHIP_SUMO2_9642 0x9642
+#define PCI_CHIP_SUMO2_9643 0x9643
+#define PCI_CHIP_SUMO2_9644 0x9644
+#define PCI_CHIP_SUMO2_9645 0x9645
+#define PCI_CHIP_SUMO_9647 0x9647
+#define PCI_CHIP_SUMO_9648 0x9648
+#define PCI_CHIP_SUMO_964A 0x964A
+#define PCI_CHIP_SUMO_964E 0x964E
+#define PCI_CHIP_SUMO_964F 0x964F
 #define PCI_CHIP_RS880_9710 0x9710
 #define PCI_CHIP_RS880_9711 0x9711
 #define PCI_CHIP_RS880_9712 0x9712
@@ -461,6 +472,8 @@
 #define PCI_CHIP_PALM_9803 0x9803
 #define PCI_CHIP_PALM_9804 0x9804
 #define PCI_CHIP_PALM_9805 0x9805
+#define PCI_CHIP_PALM_9806 0x9806
+#define PCI_CHIP_PALM_9807 0x9807
 #define PCI_CHIP_CYPRESS_6880 0x6880
 #define PCI_CHIP_CYPRESS_6888 0x6888
 #define PCI_CHIP_CYPRESS_6889 0x6889
@@ -469,6 +482,7 @@
 #define PCI_CHIP_CYPRESS_688D 0x688D
 #define PCI_CHIP_CYPRESS_6898 0x6898
 #define PCI_CHIP_CYPRESS_6899 0x6899
+#define PCI_CHIP_CYPRESS_689B 0x689B
 #define PCI_CHIP_CYPRESS_689E 0x689E
 #define PCI_CHIP_HEMLOCK_689C 0x689C
 #define PCI_CHIP_HEMLOCK_689D 0x689D
@@ -479,7 +493,9 @@
 #define PCI_CHIP_JUNIPER_68B0 0x68B0
 #define PCI_CHIP_JUNIPER_68B8 0x68B8
 #define PCI_CHIP_JUNIPER_68B9 0x68B9
+#define PCI_CHIP_JUNIPER_68BA 0x68BA
 #define PCI_CHIP_JUNIPER_68BE 0x68BE
+#define PCI_CHIP_JUNIPER_68BF 0x68BF
 #define PCI_CHIP_REDWOOD_68C0 0x68C0
 #define PCI_CHIP_REDWOOD_68C1 0x68C1
 #define PCI_CHIP_REDWOOD_68C7 0x68C7
@@ -500,6 +516,21 @@
 #define PCI_CHIP_CEDAR_68F8 0x68F8
 #define PCI_CHIP_CEDAR_68F9 0x68F9
 #define PCI_CHIP_CEDAR_68FE 0x68FE
+#define PCI_CHIP_CAYMAN_6700 0x6700
+#define PCI_CHIP_CAYMAN_6701 0x6701
+#define PCI_CHIP_CAYMAN_6702 0x6702
+#define PCI_CHIP_CAYMAN_6703 0x6703
+#define PCI_CHIP_CAYMAN_6704 0x6704
+#define PCI_CHIP_CAYMAN_6705 0x6705
+#define PCI_CHIP_CAYMAN_6706 0x6706
+#define PCI_CHIP_CAYMAN_6707 0x6707
+#define PCI_CHIP_CAYMAN_6708 0x6708
+#define PCI_CHIP_CAYMAN_6709 0x6709
+#define PCI_CHIP_CAYMAN_6718 0x6718
+#define PCI_CHIP_CAYMAN_6719 0x6719
+#define PCI_CHIP_CAYMAN_671C 0x671C
+#define PCI_CHIP_CAYMAN_671D 0x671D
+#define PCI_CHIP_CAYMAN_671F 0x671F
 #define PCI_CHIP_BARTS_6720 0x6720
 #define PCI_CHIP_BARTS_6721 0x6721
 #define PCI_CHIP_BARTS_6722 0x6722
@@ -512,6 +543,7 @@
 #define PCI_CHIP_BARTS_6729 0x6729
 #define PCI_CHIP_BARTS_6738 0x6738
 #define PCI_CHIP_BARTS_6739 0x6739
+#define PCI_CHIP_BARTS_673E 0x673E
 #define PCI_CHIP_TURKS_6740 0x6740
 #define PCI_CHIP_TURKS_6741 0x6741
 #define PCI_CHIP_TURKS_6742 0x6742
@@ -525,6 +557,7 @@
 #define PCI_CHIP_TURKS_6750 0x6750
 #define PCI_CHIP_TURKS_6758 0x6758
 #define PCI_CHIP_TURKS_6759 0x6759
+#define PCI_CHIP_TURKS_675F 0x675F
 #define PCI_CHIP_CAICOS_6760 0x6760
 #define PCI_CHIP_CAICOS_6761 0x6761
 #define PCI_CHIP_CAICOS_6762 0x6762
@@ -535,4 +568,5 @@
 #define PCI_CHIP_CAICOS_6767 0x6767
 #define PCI_CHIP_CAICOS_6768 0x6768
 #define PCI_CHIP_CAICOS_6770 0x6770
+#define PCI_CHIP_CAICOS_6778 0x6778
 #define PCI_CHIP_CAICOS_6779 0x6779
