@@ -34,7 +34,7 @@
 #include "xf86.h"
 #include "xf86_OSproc.h"
 
-#include "xf86PciInfo.h"
+#include "atipciids.h"
 #include "radeon.h"
 #include "radeon_reg.h"
 #include "radeon_macros.h"
